from __future__ import absolute_import
from datetime import datetime
from types import CodeType
from typing import Dict, TextIO
import uuid

from typing_extensions import TypedDict

from wsgi_lineprof.extensions import LineTiming


CodeTiming = Dict[int, LineTiming]
Measurement = Dict[CodeType, CodeTiming]
RequestMeasurement = TypedDict("RequestMeasurement", {
    "id": uuid.UUID,
    "started_at": datetime,
    "elapsed": float,
    "unit": float,
    "results": Measurement,
    "request_method": str,
    "path_info": str,
    "query_string": str,
})

# TODO: Improve definition
Stream = TextIO



from __future__ import absolute_import
from types import CodeType
from typing import Callable, Iterable, Union

from wsgi_lineprof.filters import BaseFilter
from wsgi_lineprof.types import CodeTiming, Measurement, RequestMeasurement


class LineProfilerStat(object):
    def __init__(self,
                 code,  # type: CodeType
                 timings  # type: CodeTiming
                 ):
        # type: (...) -> None
        self.code = code
        self.timings = timings
        self.total_time = sum(t.total_time for t in timings.values())

    @property
    def filename(self):
        # type: () -> str
        # TODO: Improve how to handle empty filename
        filename = self.code.co_filename
        return filename if filename is not None else ""

    @property
    def name(self):
        # type: () -> str
        return self.code.co_name

    @property
    def firstlineno(self):
        # type: () -> int
        return self.code.co_firstlineno


CallableFilterType = Callable[[Iterable[LineProfilerStat]],
                              Iterable[LineProfilerStat]]
FilterType = Union[CallableFilterType, BaseFilter]


class LineProfilerStats(object):
    def __init__(self, stats, unit):
        # type: (Iterable[LineProfilerStat], float) -> None
        self.stats = stats
        self.unit = unit  # seconds/hit

    @classmethod
    def from_request_measurement(cls, request_measurement):
        # type: (RequestMeasurement) -> LineProfilerStats
        return cls.from_measurement_and_unit(
            request_measurement["results"], request_measurement["unit"])

    @classmethod
    def from_measurement_and_unit(cls, measurement, unit):
        # type: (Measurement, float) -> LineProfilerStats
        return cls([LineProfilerStat(c, t) for c, t in measurement.items()], unit)

    def filter(self, f):
        # type: (FilterType) -> LineProfilerStats
        if isinstance(f, BaseFilter):
            return LineProfilerStats(f.filter(self.stats), self.unit)
        else:
            return LineProfilerStats(f(self.stats), self.unit)



from __future__ import absolute_import
import re
from six import StringIO
from typing import Any, Dict, Iterable, TYPE_CHECKING
from uuid import UUID

from jinja2 import Environment, PackageLoader

from wsgi_lineprof.formatter import TextFormatter
from wsgi_lineprof.stats import FilterType, LineProfilerStats
from wsgi_lineprof.types import RequestMeasurement
from wsgi_lineprof.writers import SyncStreamWriter

if TYPE_CHECKING:
    from wsgiref.types import StartResponse, WSGIEnvironment


UUID_RE = re.compile('^[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$',
                     re.I)


class ResultsApp(object):
    """WSGI application to show the profiling results"""

    def __init__(self, endpoint, results, filters):
        # type: (str, Dict[UUID, RequestMeasurement], Iterable[FilterType]) -> None
        assert endpoint.startswith("/")
        if not endpoint.endswith("/"):
            endpoint += "/"
        self.endpoint = endpoint
        self.results = results
        self.filters = filters
        self.template_env = Environment(
            loader=PackageLoader("wsgi_lineprof", "templates"), autoescape=True)

    def __call__(self, env, start_response):
        # type: (WSGIEnvironment, StartResponse) -> Iterable[bytes]
        if not self.should_handle_request(env):
            return _return_404(start_response)

        path = env["PATH_INFO"][len(self.endpoint):]
        if path == "":
            return self._handle_index(start_response)
        elif UUID_RE.match(path):
            return self._handle_detail(start_response, UUID(path))
        else:
            return _return_404(start_response)

    def should_handle_request(self, env):
        # type: (WSGIEnvironment) -> bool
        """Returns true only if this app should handle the request"""
        return bool(env["PATH_INFO"].startswith(self.endpoint))

    def _handle_index(self, start_response):
        # type: (StartResponse) -> Iterable[bytes]
        template = self.template_env.get_template("index.html")
        # To suppress the following mypy error on Python 3:
        # error: No overload variant of "reversed" matches argument
        results = self.results.values()  # type: Any
        start_response("200 OK", [("Content-Type", "text/html; charset=utf-8")])
        return [template.render(results=reversed(results)).encode("utf-8")]

    def _handle_detail(self, start_response, request_id):
        # type: (StartResponse, UUID) -> Iterable[bytes]
        request_measurement = self.results.get(request_id)
        if request_measurement is None:
            return _return_404(start_response)

        template = self.template_env.get_template("detail.html")
        stream = StringIO()  # type: Any
        writer = SyncStreamWriter(stream, TextFormatter(color=False))
        stats = LineProfilerStats.from_request_measurement(request_measurement)
        for f in self.filters:
            stats = stats.filter(f)
        writer.write(stats)
        start_response("200 OK", [("Content-Type", "text/html; charset=utf-8")])
        return [
            template.render(result=request_measurement,
                            stats=stream.getvalue()).encode("utf-8")
        ]


def _return_404(start_response):
    # type: (StartResponse) -> Iterable[bytes]
    start_response("404 Not Found", [])
    return []



from __future__ import division
from abc import ABCMeta, abstractmethod
import inspect
import itertools
import linecache
from os import path
from six import add_metaclass
from typing import Any, cast, Dict, Sequence

import colorama

from wsgi_lineprof.stats import LineProfilerStat, LineProfilerStats
from wsgi_lineprof.types import Stream


@add_metaclass(ABCMeta)
class BaseFormatter(object):
    def __init__(self,
                 *kwargs  # type: Any
                 ):
        # type: (...) -> None
        return

    @abstractmethod
    def format_stats(self, stats, stream):
        # type: (LineProfilerStats, Stream) -> None
        return


class TextFormatter(BaseFormatter):
    def __init__(self, color=False):
        # type: (bool) -> None
        self.color = color

    def format_stats(self, stats, stream):
        # type: (LineProfilerStats, Stream) -> None
        unit = stats.unit
        stream.write("Time unit: %s [sec]\n\n" % unit)
        for stat in stats.stats:
            self.format_stat(stat, stream, unit)

    def format_stat(self, stat, stream, unit):
        # type: (LineProfilerStat, Stream, float) -> None
        stream.write("File: %s\n" % stat.filename)
        stream.write("Name: %s\n" % stat.name)
        total_time = stat.total_time * unit
        stream.write("Total time: %g [sec]\n" % total_time)
        if not path.exists(stat.filename):
            # e.g., filename is <frozen importlib._bootstrap>
            stream.write("WARNING: Cannot find a file\n")
            return

        linecache.clearcache()
        lines = linecache.getlines(stat.filename)  # type: Sequence[str]
        if stat.name != "<module>":
            lines = inspect.getblock(lines[stat.firstlineno - 1:])

        template = '%6s %9s %12s %8s %7s  %-s'
        header = template % ("Line", "Hits", "Time", "Per Hit", "% Time",
                             "Code")
        stream.write(header)
        stream.write("\n")
        stream.write("=" * len(header))
        stream.write("\n")

        d = {}  # type: Dict[int, Dict[str, Any]]
        for i, code in zip(itertools.count(stat.firstlineno), lines):
            timing = stat.timings.get(i)
            if timing is None:
                d[i] = {
                    "hits": "",
                    "time": "",
                    "per_hit": "",
                    "percent": "",
                    "code": code,
                    "style": self.style_for_percent(0),
                }
            else:
                if stat.total_time == 0:
                    # TODO: Consider a better way to handle when total_time is 0
                    percent = 0.0
                else:
                    percent = 100 * timing.total_time / stat.total_time
                d[i] = {
                    "hits": timing.n_hits,
                    "time": timing.total_time,
                    "per_hit": '%.1f' % (timing.total_time / timing.n_hits),
                    "percent": '%.1f' % percent,
                    "code": code,
                    "style": self.style_for_percent(percent),
                }
        if self.color:
            colorama.init()
        for i in sorted(d.keys()):
            r = d[i]
            if self.color:
                stream.write(r["style"])
            stream.write(template % (i, r["hits"], r["time"], r["per_hit"],
                                     r["percent"], r["code"]))
        if self.color:
            stream.write(colorama.Style.RESET_ALL)
            colorama.deinit()
        stream.write("\n")

    # TODO: Make constants (percent/color) configurable
    def style_for_percent(self, percent):
        # type: (float) -> str
        """Returns ANSI style for a given percent"""
        if percent < 0.2:
            return cast(str, colorama.Fore.LIGHTBLACK_EX)
        elif percent >= 50:
            return cast(str, colorama.Fore.RED)
        elif percent >= 5:
            return cast(str, colorama.Fore.YELLOW)
        else:
            return cast(str, colorama.Fore.WHITE)



from __future__ import absolute_import
import atexit
from collections import OrderedDict
from datetime import datetime
from operator import itemgetter
from six.moves import reduce
import sys
import threading
from typing import Any, Dict, Iterable, Optional, Type, TYPE_CHECKING
import uuid

from pytz import utc

from wsgi_lineprof.app import ResultsApp
from wsgi_lineprof.formatter import TextFormatter
from wsgi_lineprof.profiler import LineProfiler
from wsgi_lineprof.stats import FilterType, LineProfilerStats
from wsgi_lineprof.types import CodeTiming, Measurement, RequestMeasurement, Stream
from wsgi_lineprof.writers import AsyncStreamWriter, BaseStreamWriter, SyncStreamWriter


if TYPE_CHECKING:
    from wsgiref.types import StartResponse, WSGIApplication, WSGIEnvironment


class LineProfilerMiddleware(object):
    def __init__(self,
                 app,  # type: WSGIApplication
                 stream=None,  # type: Optional[Stream]
                 filters=tuple(),  # type: Iterable[FilterType]
                 async_stream=False,  # type: bool
                 accumulate=False,  # type: bool
                 color=True,  # type: bool
                 profiler_class=LineProfiler,  # type: Type[LineProfiler]
                 endpoint="/wsgi_lineprof/",  # type: str
                 ):
        # type: (...) -> None
        self.app = app
        self.profiler_class = profiler_class
        # A hack to suppress unexpected mypy error on Python 2
        # error: Incompatible types in assignment
        # (expression has type "object", variable has type "TextIO")
        stdout = sys.stdout  # type: Any
        stream = stdout if stream is None else stream
        self.filters = filters
        self.accumulate = accumulate
        # TODO: Use typing.OrderedDict
        self.results = OrderedDict()  # type: Dict[uuid.UUID, RequestMeasurement]
        # Enable colorization only for stdout/stderr
        color = color and stream in {sys.stdout, sys.stderr}
        formatter = TextFormatter(color=color)
        # A lock to avoid multiple threads try to write the result at the same time
        self.writer_lock = threading.Lock()
        # Cannot use AsyncWriter with atexit
        if async_stream and not accumulate:
            self.writer = AsyncStreamWriter(stream, formatter)  # type: BaseStreamWriter
        else:
            self.writer = SyncStreamWriter(stream, formatter)
        if accumulate:
            atexit.register(self._write_result_at_exit)

        self.results_app = ResultsApp(
            endpoint=endpoint,
            results=self.results,
            filters=self.filters,
        )

    def _write_result_to_stream(self, measurement):
        # type: (Measurement) -> None
        stats = LineProfilerStats.from_measurement_and_unit(
            measurement, self.profiler_class.get_unit())
        for f in self.filters:
            stats = stats.filter(f)
        with self.writer_lock:
            self.writer.write(stats)

    def _write_result_at_exit(self):
        # type: () -> None
        """Write profiling results to the stream when exiting

        This method must be registered by only one thread.
        """
        measurement = reduce(
            _merge_results,
            map(itemgetter("results"), self.results.values()),
            {})  # type: Measurement
        self._write_result_to_stream(measurement)

    def __call__(self, env, start_response):
        # type: (WSGIEnvironment, StartResponse) -> Iterable[bytes]
        """Wrap an WSGI app with profiler

        Note that this thread may be called by multiple different threads.
        LineProfiler is not a thread-safe class.

        1. Create a profiler for this thread
        2. Run an WSGI application with profiling enabled
        3. Store the result of profiling or write it immediately
        4. Return the response from the WSGI application
        """
        if self.results_app.should_handle_request(env):
            return self.results_app(env, start_response)

        profiler = self.profiler_class()
        started_at = datetime.now(tz=utc)
        relative_start = profiler.get_timer()
        try:
            profiler.enable()
            response = self.app(env, start_response)
        finally:
            profiler.disable()
            unit = profiler.get_unit()
            elapsed = (profiler.get_timer() - relative_start) * unit

        request_id = uuid.uuid4()
        self.results[request_id] = {
            "id": request_id,
            "started_at": started_at,
            "elapsed": elapsed,
            "unit": unit,
            "results": profiler.results,
            "request_method": env["REQUEST_METHOD"],
            "path_info": env["PATH_INFO"],
            "query_string": env["QUERY_STRING"],
        }

        if not self.accumulate:
            self._write_result_to_stream(profiler.results)

        return response


def _merge_timings(a, b):
    # type: (CodeTiming, CodeTiming) -> CodeTiming
    for line, timing in b.items():
        if line in a:
            a[line] += timing
        else:
            a[line] = timing
    return a


def _merge_results(a, b):
    # type: (Measurement, Measurement) -> Measurement
    for code, timings in b.items():
        if code in a:
            a[code] = _merge_timings(a[code], timings)
        else:
            a[code] = timings
    return a



from abc import ABCMeta, abstractmethod
from six import add_metaclass
from six.moves.queue import Queue
from threading import Thread
from typing import Any, Optional

from wsgi_lineprof.formatter import BaseFormatter
from wsgi_lineprof.stats import LineProfilerStats
from wsgi_lineprof.types import Stream


@add_metaclass(ABCMeta)
class BaseStreamWriter(object):
    def __init__(self,
                 stream,  # type: Stream
                 formatter,  # type: BaseFormatter
                 *kwargs  # type: Any
                 ):
        # type: (...) -> None
        self.stream = stream
        self.formatter = formatter

    @abstractmethod
    def write(self, stats):
        # type: (LineProfilerStats) -> None
        return


class SyncStreamWriter(BaseStreamWriter):
    def write(self, stats):
        # type: (LineProfilerStats) -> None
        self.formatter.format_stats(stats, self.stream)


class AsyncStreamWriter(BaseStreamWriter):
    def __init__(self,
                 stream,  # type: Stream
                 formatter  # type: BaseFormatter
                 ):
        # type: (...) -> None
        super(AsyncStreamWriter, self).__init__(stream, formatter)
        # None in the queue is a special message to stop the writer thread
        self.queue = Queue()  # type: Queue[Optional[LineProfilerStats]]
        self.writer_thread = Thread(target=self._write)
        self.writer_thread.setDaemon(True)
        self.writer_thread.start()

    def write(self, stats):
        # type: (LineProfilerStats) -> None
        # Avoid accidentally stopping the writer thread by writing None
        if stats is None:
            return
        self.queue.put(stats)

    def _write(self):
        # type: () -> None
        """Method to run in the writer thread"""
        while True:
            stats = self.queue.get()
            if stats is None:
                return
            self.formatter.format_stats(stats, self.stream)

    def _join(self):
        # type: () -> None
        """Utility method to stop the writer thread

        Mainly for testing.
        """
        self.queue.put(None)
        self.writer_thread.join()
